unit GeoCompute;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, GeoEllipsoids, Math;

  type
    TZoneHemisphere = (hemEast, hemWest, hemNorth, hemSouth);

    TUTMProjection = record
      K0 : double;
      FE : double;
      FN : double;
      ZoneNo : integer;
      CM : double;
      LatHem : TZoneHemisphere;
      Longhem : TZoneHemisphere;
    end;

    TCoordinate = record
      X : double;
      Y : double;
    end;

    TGeo2UTM = class
    private
      FEllipsoid : TEllipsoid;
      FGeoCoor : TCoordinate;
      FUTMCoor : TCoordinate;
      FUTMProj : TUTMProjection;
      procedure SetGeoCoordinate (const geocoor : TCoordinate);
      procedure SetEllipsoid(const ellipse : TEllipsoid);
    public
      property Ellipsoid : TEllipsoid write SetEllipsoid;
      property GeoCoordinate : TCoordinate write SetGeoCoordinate; //input
      property UTMCoordinate : TCoordinate read FUTMCoor;  //output
      property UTMProjection : TUTMProjection read FUTMProj; //output
      procedure Compute;
      constructor Create;
      destructor Destroy; override;
    end;

    //On UTM Grid coordianates we need to know the zone no, latitude hemisphere and
    //longitude hemisphere also. Then we can convert utm coordinates to geographic
    //coordinate (latitude/longitude).
    TUTM2Geo = class
    private
      FEllipsoid : TEllipsoid;
      FGeoCoor : TCoordinate;
      FUTMCoor : TCoordinate;
      FUTMProj : TUTMProjection;
      procedure SetUTMProjection (const utmproj : TUTMProjection);
      procedure SetUTMCoordinate (const utmcoor : TCoordinate);
      procedure SetEllipsoid(const Ellipsoid : TEllipsoid);
    public
      property UTMProjection : TUTMProjection read FUTMProj write SetUTMProjection;
      property Ellipsoid : TEllipsoid write SetEllipsoid;
      property UTMCoordinate : TCoordinate write SetUTMCoordinate;  //input
      property GeoCoordinate : TCoordinate read FGeoCoor; //output
      procedure Compute;
      constructor Create;
      destructor Destroy; override;
    end;

implementation

//TGeo2UTM
procedure TGeo2UTM.SetEllipsoid(const ellipse : TEllipsoid);
begin
  FEllipsoid.EllipsoidName := ellipse.EllipsoidName;
  FEllipsoid.MajorAxis := ellipse.MajorAxis;
  FEllipsoid.InvFlattening := ellipse.InvFlattening;
end;

procedure TGeo2UTM.SetGeoCoordinate(const geocoor : TCoordinate);
var
  dLat, dLong : double;
begin
  FGeoCoor := geocoor;
  dLat := FGeoCoor.Y;
  dLong := FGeoCoor.X;

  //set hemisphere
  if (dLat >=0) then
    FUTMProj.LatHem := hemNorth
  else
    FUTMProj.LatHem := hemSouth;

  if (dLong >=0) then
    FUTMProj.LongHem := hemEast
  else
    FUTMProj.LongHem := hemWest;

  //compute zone
   if (FUTMProj.LongHem = hemWest) then
     FUTMProj.ZoneNo := trunc((180 + dLong) / 6) + 1
   else if (FUTMProj.LongHem = hemEast) then
     FUTMProj.ZoneNo := trunc(dLong/6) + 31;

  //compute zone
   if (FUTMProj.LatHem = hemNorth) then
     FUTMProj.FN := 0
   else if (FUTMProj.LatHem = hemSouth) then
     FUTMProj.FN := 10000000;

   FUTMProj.CM := 6 * FUTMProj.ZoneNo - 183;
end;

procedure TGeo2UTM.Compute;
var
  K0, FE, FN : extended;
  lat, long : extended; //in radian.
  a, b, f, e, et2, rm, n, rho, nu, p, cm : extended;
  M, A0, B0, C0, D0, E0 : extended;
  Ki, Kii, Kiii, Kiv, Kv, A6 : extended;

begin
  //Assign to new variable for easy looking at mathematic formula.
  K0 := FUTMProj.K0;
  FE := FUTMProj.FE;
  FN := FUTMProj.FN;
  a := FEllipsoid.MajorAxis;
  f := 1 / FEllipsoid.InvFlattening;
  long := PI/180 * FGeoCoor.X; //convert latitude to radian.
  lat := PI/180 * FGeoCoor.Y;
  cm := PI/180 * FUTMProj.CM;
  p := long - cm; //already in radian.

  b := a * (1 - f); //Minor axis;

  // =.08 approximately. This is the eccentricity of the earth's elliptical crosssection.
  e := sqrt(1 - b*b / (a*a));

  //= .007 approximately. The quantity e' only occurs in even powers so it need only be calculated as e'2
  et2 := power((e * a/b), 2);

  n := (a - b) / (a + b);

  //This is the radius of curvature of the earth in the meridian plane.
  rho := a * (1 - e*e) / power (1 - power(e*sin(lat),2), 3/2);

  //This is the radius of curvature of the earth perpendicular to the meridian plane.
  nu := a / sqrt(1 - sqr(e*sin(lat)));

  //=================== compute Meridian Arc===================================
  //A' = a(1 - n + (5/4)(n2- n3) + (81/64)(n4- n5))
  A0 := a * (1 - n + (5.0/4.0)*(n*n - n*n*n) + (81.0/64.0)*(power(n,4) - power(n,5)));
  //B' = (3an/2)[1 - n + (7/8)(n2- n3) + (55/64)(n4- n5) ...]
  B0 := (3*a*n/2) * (1 - n + (7/8)*(n*n - n*n*n) + (55/64)* power(n,4) - power(n,5));
  //C' = (15 tan2/16)[1 - n + (3/4)(n2- n3) ...]
  C0 := (15*a*n*n/16) * (1 - n + (3/4)*(n*n - n*n*n));
  //D' = (35 tan3/48)[1 - n + (11/16)(n2- n3) ...]
  D0 := (35*a*n*n*n/48) * (1 - n + (11/16)*(n*n - n*n*n));
  //E' = (315 tan4/512)[1 - n ...]
  E0 := (315*a*power(n,4)/512) * (1-n);
  //M = A'lat - B'sin(2lat) + C'sin(4lat) - D'sin(6lat) + E'sin(8lat).
  M := A0*lat - B0*sin(2*lat) + C0*sin(4*lat) - D0*sin(6*lat) + E0*sin(8*lat);
  //===========================================================================

  //===================Converting Lat/Long to UTM Grid Coordinate==============
  Ki := M * K0;
  Kii := K0 * nu * sin(2*lat) / 4;
  Kiii := (K0 * nu * sin(lat) * power(cos(lat),3) / 24) * ((5 - tan(lat)*tan(lat)
           + 9*et2 * cos(lat)*cos(lat) + 4 * et2*et2 * power(cos(lat),4)));
  Kiv := K0 * nu * cos(lat);
  Kv := (K0 * nu * power(cos(lat),3)/6) * (1 - tan(lat)*tan(lat) + et2*cos(lat)*cos(lat));
  //Now we 'v got Grid UTM Coordinates.
  FUTMCoor.Y := FN + Ki + Kii*p*p + Kiii*power(p,4);
  FUTMCoor.X := FE + Kiv*p + Kv*p*p*p;
end;

constructor TGeo2UTM.Create;
begin
  inherited create;
  FUTMProj.K0 := 0.9996;
  FUTMProj.FE := 500000;
  FEllipsoid := TEllipsoid.Create;
end;

destructor TGeo2UTM.Destroy;
begin
  FEllipsoid.Free;
  inherited Destroy;
end;


//TUTM2Geo
procedure TUTM2Geo.SetEllipsoid(const Ellipsoid : TEllipsoid);
begin
  FEllipsoid.EllipsoidName := Ellipsoid.EllipsoidName;
  FEllipsoid.MajorAxis := Ellipsoid.MajorAxis;
  FEllipsoid.InvFlattening := Ellipsoid.InvFlattening;
end;

procedure TUTM2Geo.SetUTMProjection(const utmproj : TUTMProjection);
begin
   FUTMProj.LongHem := utmproj.LongHem;
   FUTMProj.LatHem := utmproj.LatHem;
   if (utmproj.LatHem = hemNorth) then
     FUTMProj.FN := 0
   else if (utmproj.LatHem = hemSouth) then
     FUTMProj.FN := 10000000;

   FUTMProj.CM := 6 * utmproj.ZoneNo - 183;

   if (FUTMProj.CM < 0) then
     FUTMProj.Longhem := hemWest
   else
     FUTMProj.Longhem := hemEast;

end;

procedure TUTM2Geo.SetUTMCoordinate(const utmcoor : TCoordinate);
var
  dN, dE : double;
begin
  FUTMCoor := utmcoor;
  dN := FGeoCoor.Y;
  dE := FGeoCoor.X;
end;

procedure TUTM2Geo.Compute;
var
  K0, FE, FN, a, b, f, e, et2, cm : double;
  mu, M, e1 : double;
  J1, J2, J3, J4 : double;
  C1, T1, R1, N1, D : double;
  fp, Q1, Q2, Q3, Q4, Q5, Q6, Q7 : double;
  lat, long : double;
  x, y : double;
begin
  //Assign to new variable for easy looking at mathematic formula.
  K0 := FUTMProj.K0;
  FE := FUTMProj.FE;
  FN := FUTMProj.FN;
  a := FEllipsoid.MajorAxis;
  f := 1 / FEllipsoid.InvFlattening;
  x := (FE - FUTMCoor.X);
  if (FUTMProj.LatHem = hemNorth) then
    y := FUTMCoor.Y
  else if (FUTMProj.LatHem = hemSouth) then
    y := (10000000 - FUTMCoor.Y);  //temporary

  cm := PI/180 * FUTMProj.CM;
  b := a * (1 - f); //Minor axis;

  // =.08 approximately. This is the eccentricity of the earth's elliptical crosssection.
  e := sqrt(1 - b*b / (a*a));

  //= .007 approximately. The quantity e' only occurs in even powers so it need only be calculated as e'2
  et2 := power((e * a/b), 2);

  M := y / K0; // M is Meridian Arc;
  //Compute footprint Latitude.
  //mu = M/[a(1 - e2/4 - 3e4/64 - 5e6/256...)
  mu := M/(a*(1 - e*e/4 - 3/64*power(e,4) - 5/256*power(e,6)));
  //e1 = [1 - (1 - e2)1/2]/[1 + (1 - e2)1/2]
  e1 := (1 - sqrt(1 - e*e))/(1 + sqrt(1 - e*e));
  //J1 = (3e1/2 - 27e13/32 ..)
  J1 := (3*e1/2 - 27*e1*e1*e1/32);
  //J2 = (21e12/16 - 55e14/32 ..)
  J2 := 21*e1*e1/16 - 55*power(e1,4)/32;
  //J3=(151e13/96 ..)
  J3 := 151*e1*e1*e1/96;
  //J4 = (1097e14/512 ..)
  J4 := 1097*power(e1,4)/512;
  //Footprint Latitude fp = mu + J1sin(2mu) + J2sin(4mu) + J3sin(6mu) + J4sin(8mu)
  fp := mu + J1*sin(2*mu) + J2*sin(4*mu) + J3*sin(6*mu) + J4 * sin(8*mu);


  //C1=e'2cos2(fp)
  C1 := et2*cos(fp)*cos(fp);
  //T1 = tan2(fp)
  T1 := tan(fp)*tan(fp);
  //R1 = a(1-e2)/(1-e2sin2(fp))3/2
  R1 := a*(1-e*e)/power(1-e*e*sin(fp)*sin(fp), 1.5);
  //N1 = a/(1-e2sin2(fp))1/2
  N1 := a / sqrt(1 -e*e*sin(fp)*sin(fp));
  //D = x/(N1k0)
  D := x / (N1 * K0);
  //Compute Latitude
  //Q1 = N1 tan(fp)/R1
  Q1 := N1 * tan(fp)/R1;
  //Q2 = (D2/2)
  Q2 := D*D/2;
  //Q3 = (5 + 3T1 + 10C1 - 4C12-9e'2)D4/24
  Q3 := (5 + 3*T1 + 10*C1 - 4*C1*C1-9*et2)*power(D,4)/24;
  //Q4 = (61 + 90T1 + 298C1 +45T12- 3C12-252e'2)D6/720
  Q4 := (61 + 90*T1 + 298*C1 + 45*T1*T1 - 3*C1*C1 - 252*et2)* power(D,6)/720;


  Q5 := D;
  //Q6 = (1 + 2T1 + C1)D3/6
  Q6 := (1 + 2*T1 + C1) *D*D*D/6;
  //Q7 = (5 - 2C1 + 28T1 - 3C12+ 8e'2+ 24T12)D5/120
  Q7 := (5 - 2*C1 + 28*T1 - 3*C1*C1 + 8*et2 + 24*T1*T1)*power(D,5)/120;

  //lat
  lat := 180/PI *(fp - Q1*(Q2 + Q3 + Q4));

  //long = long0 + (Q5 - Q6 + Q7)/cos(fp)
  long := 180/PI *(CM - (Q5 - Q6 + Q7)/cos(fp));
  FGeoCoor.Y := lat;
  FGeoCoor.X := long;
end;

constructor TUTM2Geo.Create;
begin
  inherited create;
  FEllipsoid := TEllipsoid.Create;
  FUTMProj.K0 := 0.9996;
  FUTMProj.FE := 500000;
end;

destructor TUTM2Geo.Destroy;
begin
  FEllipsoid.Free;
  inherited Destroy;
end;

end.
                                      .
